;;; cps.ss
;;; Copyright (c) 2005, R. Kent Dybvig, Simon L. Peyton Jones, and Amr Sabry

;;; Naive CPS converter for Chez Scheme expander output, with support
;;; for primitives and top-level variables.

;;; Known primitive calls are handled inline.  For non-call references to
;;; known primitives, the variable $cps-<primname> is defined to a run-time
;;; generated cps version of the primitive.

(define ($cps-prim $cps-x v)
  (cond
    [(top-level-bound? $cps-x) (top-level-value $cps-x)]
    [else
     (let ([cps-v (lambda (*k . args) (*k (apply v args)))])
       (define-top-level-value $cps-x cps-v)
       cps-v)]))

;;; Top-level variable references are handled similarly.  If the variable
;;; cps-varname is not already bound (e.g., via set! or an as a side
;;; effect of an earlier reference) and the variable varname is bound to
;;; a procedure, the variable cps-varname is defined to a run-time generated
;;; cps version of varname's top-level value.

(define ($cps-global x cps-x)
  (cond
    [(top-level-bound? cps-x) (top-level-value cps-x)]
    [(and (top-level-bound? x)
          (let ([v (top-level-value x)])
            (and (procedure? v) v))) =>
     (lambda (v)
       (let ([cps-v (lambda (*k . args) (*k (apply v args)))])
         (define-top-level-value cps-x cps-v)
         cps-v))]
    [else (error #f "variable ~s (as ~s) is not bound" x cps-x)]))

;;; cps converts an expanded expression e into a lambda expression with
;;; one argument, the initial continuation k0.  For example,
;;; (cps '3) -> (lambda (k) (k 3)).

(define (cps x)
  (define-syntax match ; rudimentary matcher
    (lambda (x)
      (define match-pat
        (lambda (t pat fail succ)
          (with-syntax ((t t) (fail fail) (succ succ))
            (syntax-case pat (?)
              (id (identifier? #'id) #'(let ((id t)) succ))
              ((? stuff id)
               (identifier? #'id)
               #'(if (let ((p stuff))
                       (if (procedure? p)
                           (p t)
                           (eqv? t p)))
                     (let ((id t)) succ)
                     fail))
              ((? stuff) (match-pat #'t #'(? stuff x) #'fail #'succ))
              (() #'(if (null? t) succ fail))
              ((p1 . p2)
               (with-syntax (((tcar tcdr) (generate-temporaries '(tcar tcdr))))
                 (with-syntax ((succ (match-pat #'tcar #'p1 #'fail
                                       (match-pat #'tcdr #'p2 #'fail
                                         #'succ))))
                   #'(if (pair? t)
                         (let ((tcar (car t)) (tcdr (cdr t))) succ)
                         fail))))))))
      (syntax-case x (else)
        [(_ x (else e1 e2 ...)) #'(begin x e1 e2 ...)]
        [(_ x (pat e1 e2 ...) c ...)
         (with-syntax ((e (match-pat #'t #'pat #'(fail)
                            #'(begin e1 e2 ...))))
           #'(let ((t x))
               (let ((fail (lambda () (match t c ...))))
                 e)))])))
  
  (define decorate-symbol
    (lambda (x)
      (string->symbol (string-append "cps-" (symbol->string x)))))
  (define $decorate-symbol
    (lambda (x)
      (string->symbol (string-append "$cps-" (symbol->string x)))))
  (define fappend
    (lambda (x y)
      (cond
        [(pair? x) (cons (car x) (fappend (cdr x) y))]
        [(null? x) y]
        [else (cons x y)])))
  (define symlhs
    (lambda (x r)
      (if (memq x r) x (decorate-symbol x))))
  (define symrhs
    (lambda (x r)
      (if (memq x r)
          x
          `($cps-global ',x ',(decorate-symbol x)))))
  (define prim?
    (lambda (x)
      (match x
        [((? '\$primitive) x) #t]
        [((? '\$primitive) n x) #t]
        [else #f])))
  (define *k (gensym "*k"))
  (define ktail (lambda (t) `(,*k ,t)))
  (define cps
    (lambda (x r k)
      (match x
        [((? 'quote) d) (k x)]
        [(? prim? x) (k `($cps-prim ',($decorate-symbol x) ,x))]
        [(? symbol? x) (k (symrhs x r))]
        [((? 'set!) x e)
         (cps e r
           (lambda (t)
             (k `(set! ,(symlhs x r) ,t))))]
        [((? 'let) binds . rest)
         (cps `((lambda ,(map car binds) . ,rest) . ,(map cadr binds)) r k)]
        [((? 'lambda) formals exp . rest)
         (k `(lambda (,*k . ,formals)
               ,(cps-body exp rest (fappend formals r) ktail)))]
        [(((? 'lambda) formals exp . rest) . args)
         (cps* args r
           (lambda (temps)
             `(let ,(map list formals temps)
                ,(cps-body exp rest (fappend formals r) k))))]
        [((? 'letrec) bindings exp . rest)
         (if (null? bindings)
             (cps-body exp rest r k)
             (cps (let ((formals (map car bindings))
                        (args (map cadr bindings)))
                    (let ((tvars (map (lambda (x) (gensym)) formals)))
                      `((lambda ,formals
                          (begin
                            ((lambda ,tvars
                               (begin
                                 ,@(map (lambda (x y) `(set! ,x ,y))
                                        formals
                                        tvars)))
                             ,@args)
                            ,exp
                            ,@rest))
                        ,@(map (lambda (x) '(#2%void)) formals))))
               r k))]
        [((? 'letrec*) bindings exp . rest)
         (cps (let ((formals (map car bindings))
                    (args (map cadr bindings)))
                `((lambda ,formals
                    (begin
                      ,@(map (lambda (x y) `(set! ,x ,y))
                             formals
                             args)
                      ,exp
                      ,@rest))
                  ,@(map (lambda (x) '(#2%void)) formals)))
            r k)]
        [((? 'begin) exp . rest) (cps-body exp rest r k)]
        [((? 'if) e1 e2 e3) ; replicates continuation
         (cps e1 r
           (lambda (t1)
             (if (eq? k ktail)
                 `(if ,t1 ,(cps e2 r k) ,(cps e3 r k))
                 (let ([tk (gensym)] [tv (gensym)])
                   `(let ((tk (lambda (,tv) ,(k tv))))
                      (if ,t1
                          ,(cps e2 r (lambda (t) `(tk ,t)))
                          ,(cps e3 r (lambda (t) `(tk ,t)))))))))]
        [((? prim? e0) . args)
         (cps* args r
           (lambda (temps)
             (k `(,e0 ,@temps))))]
        [((? symbol? e0) . args) (cps-call (symrhs e0 r) args r k)]
        [(e0 . args)
         (cps e0 r
           (lambda (t0)
             (cps-call t0 args r k)))]
        [else (k x)])))
  (define cps-body
    (lambda (exp rest r k)
      (let f ((exp exp) (rest rest))
        (if (null? rest)
            (cps exp r k)
            (cps exp r
              (lambda (t)
                `(begin ,t ,(f (car rest) (cdr rest)))))))))
  (define cps-call
    (lambda (t0 ls r k)
      (cps* ls r
        (lambda (temps)
          `(,t0
            ,(if (eq? k ktail)
                 *k
                 (let ((t (gensym))) `(lambda (,t) ,(k t))))
            ,@temps)))))
  (define cps*
    (lambda (ls r k)
      (let f ((ls ls) (temps '()))
        (if (null? ls)
            (k (reverse temps))
            (cps (car ls) r (lambda (t) (f (cdr ls) (cons t temps))))))))
  `(lambda (,*k) ,(cps x '() ktail)))
